# Copyright (c) 2023-2024 Salah Berra and contributors
# Distributed under the the GNU General Public License (See accompanying file LICENSE or copy
# at https://www.gnu.org/licenses/)

__author__ = "Salah Berra and contributors"
__copyright__ = "Copyright 2022-2023"
__license__ = "GNU GPL"
__version__ = "0.1.0"
__maintainer__ = "Salah Berra / Mohamed Nennouche"
__email__ = "."
__status__ = "dev"

# Note that each Alghorithm import from method module and run in the main code Iterative.
from .methods import *
from .train_methods import *
from .utils import *